#!/bin/sh

ARG=$1
DIST=2
RPM_VERSION=0.36Beta
RPM_RELEASE=1
RPM_PCARCH=i386
RPMNAME[0]=AVerMedia-MTSA-SUSE10
RPMNAME[1]=AVerMedia-MTSA-SUSE10-SMP

RESULT_FILE=".result"
INSTALL_S=0

install_check_rpm()
{
	RPM_NAME=${RPMNAME[$1]}
	RPM_FILE=$RPM_NAME-$RPM_VERSION-$RPM_RELEASE.$RPM_PCARCH.rpm

	if [[ ! -e ./$RPM_FILE ]]; then
		dialog --title "Install Error" \
		--backtitle "AVerMedia Device Driver" \
		--msgbox "There is no $RPM_FILE" 12 60
		exit 0
	fi
}

install_rpm()
{
	RPM_NAME=${RPMNAME[$1]}
	RPM_FILE=$RPM_NAME-$RPM_VERSION-$RPM_RELEASE.$RPM_PCARCH.rpm

	#FIXME: more and more avermedia driver installed
	#       will affect the check
	RESULT=`rpm -qa $RPM_NAME`
	if [[ "" == $RESULT ]]; then

		rpm -i $RPM_FILE 2> $RESULT_FILE
		RESULT=`cat $RESULT_FILE | grep error | wc -l`

		if [[ $RESULT -eq 0 ]]; then
			# show succeed message
		        dialog --title "Install" \
			--backtitle "AVerMedia Device Driver" \
			--msgbox "The $RPM_FILE installation is successful." 12 60
			INSTALL_S=1
		else
		        dialog --title "Install" \
			--backtitle "AVerMedia Device Driver" \
			--msgbox "The $RPM_FILE installation is failed!!\n\nmessage:\n`cat $RESULT_FILE`" 12 60
		fi
	else
		dialog --title "Install Warning" \
		--backtitle "AVerMedia Device Driver" \
		--msgbox "The $RPM_FILE is already installed.\n\nSkip it!!" 12 60
	fi

	rm -rf $RESULT_FILE
}

auto_insmod()
{
	rmmod saa7134     > /dev/null 2>&1
	modprobe saa713x  > /dev/null 2>&1

        dialog --title "Install" --backtitle "AVerMedia Device Driver" \
	--msgbox "The driver is installed.\nYou might reboot your system to enjoy it!!" 12 60

}

mdk_post_install()
{
        #To ensure Mandriva load saa713x at boot time
        if [ -e /etc/modprobe.preload ]; then
            echo saa713x >> /etc/modprobe.preload
        fi
}

suse_post_install()
{
        #To ensure Mandriva load saa713x at boot time
        if [ -e /etc/rc.d/boot.local ]; then
            echo "modprobe saa713x" >> /etc/rc.d/boot.local
        fi	
}

fc3_post_install()
{
        if [[ ! -e /etc/udev/devices/dvb ]]; then
                mkdir /etc/udev/devices/dvb >/dev/null 2>&1
                chmod 755 /etc/udev/devices/dvb >/dev/null 2>&1

                for i in `seq 0 3`; do
                        mkdir /etc/udev/devices/dvb/adapter$i >/dev/null 2>&1
                        chmod 755 /etc/udev/devices/dvb/adapter$i >/dev/null 2>&1
                        mknod -m 0660 /etc/udev/devices/dvb/adapter$i/video0    c 212   `expr 64 \* $i + 0` >/dev/null 2>&1
                        mknod -m 0660 /etc/udev/devices/dvb/adapter$i/audio0    c 212   `expr 64 \* $i + 1` >/dev/null 2>&1
                        mknod -m 0660 /etc/udev/devices/dvb/adapter$i/frontend0 c 212   `expr 64 \* $i + 3` >/dev/null 2>&1
                        mknod -m 0660 /etc/udev/devices/dvb/adapter$i/demux0    c 212   `expr 64 \* $i + 4` >/dev/null 2>&1
                        mknod -m 0660 /etc/udev/devices/dvb/adapter$i/dvr0      c 212   `expr 64 \* $i + 5` >/dev/null 2>&1
                        mknod -m 0660 /etc/udev/devices/dvb/adapter$i/ca0       c 212   `expr 64 \* $i + 6` >/dev/null 2>&1
                        mknod -m 0660 /etc/udev/devices/dvb/adapter$i/net0      c 212   `expr 64 \* $i + 7` >/dev/null 2>&1
                        mknod -m 0660 /etc/udev/devices/dvb/adapter$i/osd0      c 212   `expr 64 \* $i + 8` >/dev/null 2>&1
                        chown root.video /etc/udev/devices/dvb/adapter$i/* >/dev/null 2>&1
                done
        fi
}

install()
{
	for ((i=0 ; i<$DIST ; i++)); do
		install_check_rpm $i 
	done

	dialog --title "Install" --backtitle "AVerMedia Device Driver" \
	--yesno "This is AVerMedia Device Driver Wizard.\nDo you want to install device driver?" 12 60
	sel=$?
	case $sel in
	0) break
	;;
	1) exit 0
	;;
	esac

	for ((i=0 ; i<$DIST ; i++)); do
		install_rpm $i 
	done
	if [[ INSTALL_S -eq 1 ]]; then
		suse_post_install
		auto_insmod
	fi
}

uninstall_rpm()
{
	RPM_NAME=${RPMNAME[$1]}
	RPM_FILE=$RPM_NAME-$RPM_VERSION-$RPM_RELEASE.$RPM_PCARCH.rpm

	#FIXME: more and more avermedia driver installed
	#       will affect the check
	RESULT=`rpm -qa $RPM_NAME`
	if [[ "" == $RESULT ]]; then
                dialog --title "Uninstall Warning" \
                --backtitle "AVerMedia Device Driver" \
                --msgbox "The $RPM_FILE is not installed.\n\nSkip it!!" 12 60
	else
		rpm -e $RPM_NAME-$RPM_VERSION-$RPM_RELEASE 2> $RESULT_FILE
		RESULT=`cat $RESULT_FILE | grep error | wc -l`

		if [[ $RESULT -eq 0 ]]; then
			# show succeed message
			dialog --title "Uninstall" \
			--backtitle "AVerMedia Device Driver" \
			--msgbox "The $RPM_FILE Uninstallation is successful." 12 60
		else
		        dialog --title "UnInstall" \
			--backtitle "AVerMedia Device Driver" \
			--msgbox "The $RPM_FILE installation is failed!!\n\nmessage:\n`cat $RESULT_FILE`" 12 60
		fi
	fi
	
	rm -rf $RESULT_FILE	
}

mdk_post_uninstall()
{
        #Remove saa713x from Mandriva boot load modules
        if [ -e /etc/modprobe.preload ]; then
            cp -f /etc/modprobe.preload /etc/modprobe.preload.tmp
            cat /etc/modprobe.preload.tmp | grep -v "saa713x" > /etc/modprobe.preload
            rm -f /etc/modprobe.preload.tmp
        fi
}

suse_post_uninstall()
{
        #Remove saa713x from Mandriva boot load modules
        if [ -e /etc/rc.d/boot.local ]; then
            cp -f /etc/rc.d/boot.local /etc/rc.d/boot.local.tmp
            cat /etc/rc.d/boot.local.tmp | grep -v "modprobe saa713x" > /etc/rc.d/boot.local
            rm -f /etc/rc.d/boot.local.tmp
        fi
}

uninstall()
{
	dialog --title "Uninstall" --backtitle "AVerMedia Device Driver" \
	--yesno "This is AVerMedia Device Driver Wizard.\nDo you want to uninstall device driver?" 12 60
	sel=$?
	case $sel in
	0) break
	;;
	1) exit 0
	;;
	esac

	for ((i=0 ; i<$DIST ; i++)); do
		uninstall_rpm $i
	done

	suse_post_uninstall
}

usage()
{
	echo "Usage:   $0 [OPTION]"
	echo ""
	echo " --install      install device driver"
	echo " --uninstall    uninstall device driver"
	echo ""
}

if [[ $ARG == "" ]]; then
	usage
	exit 0
fi

case $ARG in
	"--install")
	  install
	  break
	  ;;
	"--uninstall")
	  uninstall
	  break
	  ;;
	*)
	  usage
	  exit 0
	  ;;
esac
