/*
 *  cc.c -- closed caption decoder
 *  Mike Baker (mbm@linux.com)
 *  (based on code by timecop@japan.co.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  ========================================================================
 *  vbi-cc.c is the VBI CLOSED CAPTION DECODER.
 *
 *  taken from cc.c by Mike Baker (mbm@linux.com)
 *  (based on code by timecop@japan.co.jp)
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <getopt.h>
#include <asm/types.h>
#include <sys/mman.h>
#include <X11/Xlib.h>
#include "videodev2.h"
Display *dpy;
Window Win,Root;
char dpyname[256] = "";
GC WinGC;
GC WinGC0;
GC WinGC1;
int x;

char vbi_name[256] = "/dev/vbi0";

#define MAX_BUFFER	16

//XDSdecode
char	info[8][25][256]; 
char	newinfo[8][25][256];
char	*infoptr=newinfo[0][0];
int	mode,type;
char	infochecksum;

static int	gRawline;				//which line is interesting
static struct v4l2_capability	gCap;
static int	gVAMap, gMapSize;			//where to map and size
static int	gBPL, gBufferSize;			//how many bytes each buffer
static unsigned char *gBuffer[MAX_BUFFER];
static int	gBufferCount=MAX_BUFFER;
static enum v4l2_memory gReadMethod=V4L2_MEMORY_MMAP;	//default memory map 


static unsigned long getColor(char *colorName, float dim)
{
	XColor Color;
	XWindowAttributes Attributes;

	XGetWindowAttributes(dpy, Root, &Attributes);
	Color.pixel = 0;

	XParseColor (dpy, Attributes.colormap, colorName, &Color);
	Color.red=(unsigned short)(Color.red/dim);
	Color.blue=(unsigned short)(Color.blue/dim);
	Color.green=(unsigned short)(Color.green/dim);
	Color.flags=DoRed | DoGreen | DoBlue;
	XAllocColor (dpy, Attributes.colormap, &Color);

	return Color.pixel;
}

static void raw_draw(unsigned char *buf)
{
	int x;
	XClearArea(dpy,Win,0,0, gBPL/2,128,0);
	XDrawLine(dpy,Win,WinGC1,0,128-85/2, gBPL/2,128-85/2);
	for (x=0;x<gBPL/2;x++)
		if (buf[gBPL * gRawline+x*2]/2<128 &&
					buf[gBPL * gRawline+x*2+2]/2 < 128)
			XDrawLine(dpy,Win,WinGC0,x,
				128-buf[gBPL * gRawline+x*2]/2, x+1,
				128-buf[gBPL * gRawline+x*2+2]/2);
}

static void get_param(int argc, char **argv)
{
	int arg;

	while( 1 ) {
		if( -1 == (arg = getopt(argc, argv, "?hHr:")) )
			break;

		switch( arg ) {
		case '?':
		case 'h':
		case 'H':
			printf("Usage: vbi-cc -r LINE \n\n");
			printf("\tLINE - starts from 0 of all captured lines which is interesting\n");
			printf("\t\t i.e. LINE0 is 10/7 for NTSC/PAL)\n");
			exit(1);
		case 'r':
		case 'R':
			gRawline = atoi(optarg);
			break;
		}
	}

	if(optind < argc) 
		strncpy(vbi_name, argv[optind], sizeof(vbi_name));
}

/*
 * In	: gBPL
 */
static void init_win()
{
	dpy=XOpenDisplay(dpyname);
	Root=DefaultRootWindow(dpy);
	Win = XCreateSimpleWindow(dpy, Root, 10, 10, gBPL/2,
							128, 0, 0, 0);
	WinGC = XCreateGC(dpy, Win, 0, NULL);
	WinGC0 = XCreateGC(dpy, Win, 0, NULL);
	WinGC1 = XCreateGC(dpy, Win, 0, NULL);
	XSetForeground(dpy, WinGC, getColor("blue",1));
	XSetForeground(dpy, WinGC0, getColor("green",1));
	XSetForeground(dpy, WinGC1, getColor("red",1));
	XMapWindow(dpy, Win);
}

/*
 * Allocate buffer of device but return the length for
 * mapping at once, which must be continuous
 * In	: fd - dev file
 *	  count - number of buffer requested
 * Out	: count - number of buffer allocated
 *	  dev_off - from where (of dev) we start mapping
 * Ret	: 0 - fail
 *	  otherwise - length to map from offset (of device)
 */
static int req_buffer(int fd, int *dev_off, int *count, int *buf_sz)
{
	struct v4l2_requestbuffers reqbuf;
	struct v4l2_buffer buf;
	int i, offset = -1, size;

	memset(&reqbuf, 0, sizeof(reqbuf));
	reqbuf.type = V4L2_BUF_TYPE_VBI_CAPTURE;
	reqbuf.memory = V4L2_MEMORY_MMAP;
	reqbuf.count = *count;

	if( ioctl(fd, VIDIOC_REQBUFS, &reqbuf) ) {
		perror("VIDIOC_REQBUFS");
		return 0;
	} 

	for( i=0, size=0, *count=0; i<reqbuf.count; i++ ) {
		memset(&buf, 0, sizeof(buf));
		buf.type = reqbuf.type;
		buf.index = i;
		if( ioctl(fd, VIDIOC_QUERYBUF, &buf) ) {
			perror("VIDIOC_QUERYBUF");
			return 0;
		}

		if( 0==size ) {
			offset = *dev_off = buf.m.offset;
			*buf_sz = buf.length;
		}

		if( offset==buf.m.offset ) {
			offset += buf.length;
			size += buf.length;
			*count = *count + 1;
		} else {
			/* not continuous*/
			printf("Buffers not all continuous\n");
			break;
		}
	}

	return size;
}

/*
 * Query Buffers
 * In	: gVAMap, gMapSize, gBufferCount
 * Ret	: 0 - fail
 *	  otherwise - number of buffer mapped
 */
static void buffer_q(int fd)
{
	unsigned char *map_start;
	int i;
	struct v4l2_buffer buf;

	for( i=0, map_start=(unsigned char *)gVAMap; i<gBufferCount;
		     i++) {
		memset(&buf, 0, sizeof(buf));
		buf.index = i;
		buf.type = V4L2_BUF_TYPE_VBI_CAPTURE;
		buf.memory = gReadMethod;
		if( ioctl(fd, VIDIOC_QBUF, &buf) ) {
			perror("VIDIOC_QBUF map");
			munmap((void *)gVAMap, gMapSize);
			exit(1);
		}
		gBuffer[i] = map_start;
		map_start += buf.length;
	}
}

/*
 * Check if VBI supported and get format as well
 * Out	: gCap - device capability saved
 *	  gBufferSize, gBPL update according to driver
 * Ret	: 1 - success
 *	  0 - fail
 */
static int query_fmt(int vbifd)
{
	struct v4l2_format fmt;
	v4l2_std_id std;

	/* Check if vbi supoprted */
	if( ioctl(vbifd, VIDIOC_QUERYCAP, &gCap) ) {
		perror("VIDIOC_QUERYCAP");
		return 0;
	}

	if( 0==(gCap.capabilities&V4L2_CAP_VBI_CAPTURE) ) {
		printf("VBI is not supported !\n");
		return 0;
	}

	/* get current setting */
	fmt.type = V4L2_BUF_TYPE_VBI_CAPTURE;
	if( ioctl(vbifd, VIDIOC_G_FMT, &fmt) ) {
		perror("VIDIOC_G_FMT");
		return 0;
	}

	/* get standard so we know the start and end line */
	if( ioctl(vbifd, VIDIOC_G_STD, &std) ) {
		perror("VIDIOC_G_STD");
		return 0;
	}

	if( std & V4L2_STD_NTSC ) {
		fmt.fmt.vbi.start[0] = 10;
		fmt.fmt.vbi.start[1] = fmt.fmt.vbi.start[0] - 10 + 273;
		fmt.fmt.vbi.count[0] = 21 - 10 +1;
	} else {
		fmt.fmt.vbi.start[0] = 7;
		fmt.fmt.vbi.start[1] = fmt.fmt.vbi.start[0] - 7 + 319;
		fmt.fmt.vbi.count[0] = 23 - 7 +1;
	}

	if( fmt.fmt.vbi.count[0]<=0 ) {
		printf("param error\n");
		return 0;
	} else fmt.fmt.vbi.count[1] = fmt.fmt.vbi.count[0];

	printf("To set from line %d for %d lines\n", fmt.fmt.vbi.start[0],
	       fmt.fmt.vbi.count[0]);

	if( ioctl(vbifd, VIDIOC_S_FMT, &fmt) ) {
		perror("VIOIOC_S_FMT");
		return 0;
	}

	/* assume all sample single byte */
	gBPL = fmt.fmt.vbi.samples_per_line;
	gBufferSize = fmt.fmt.vbi.count[0]+fmt.fmt.vbi.count[1];
	if( gRawline>=gBufferSize ) {
		printf("Invalid LINE #\n");
		return 0;
	}
	gBufferSize *= gBPL;

	printf("Set to BPL=%d, from line %d for %d lines", gBPL,
		fmt.fmt.vbi.start[0], fmt.fmt.vbi.count[0]);
	printf(", buffer size = %d\n", gBufferSize);
	return 1;
}

/*
 * Get vbi through memory mapping
 */
static void dma_go(int vbifd)
{
	struct v4l2_buffer buf;
	unsigned int j;

	if( ioctl(vbifd, VIDIOC_STREAMON, &j) ) {
		perror("VIDIOC_STREAMON");
		goto err;
	}


	memset(&buf, 0, sizeof(buf));
	while( 1 ) {
		buf.type = V4L2_BUF_TYPE_VBI_CAPTURE;
		if( ioctl(vbifd, VIDIOC_DQBUF, &buf) ) {
			perror("VIDIOC_DQBUF");
			goto err;
		}
		
		j = buf.index;
		raw_draw(gBuffer[j]);
		XFlush(dpy);

		if( ioctl(vbifd, VIDIOC_QBUF, &buf) ) {
			perror("VIDIOC_QBUF");
			goto err;
		}
	}

	/* something wrong */
err:
	if( ioctl(vbifd, VIDIOC_STREAMOFF, &j) )
		perror("VIDIOC_STREAMOFF");

	munmap((void *)gVAMap, gMapSize);
	close(vbifd);
}

/*
 * Out	: gBufferCount - buffer allocated or mapped
 *	  V4L2_MEMORY_MMAP -
 *		gVAMap and gMapSize set
 */
static void buffer_alloc(int vbifd)
{
	int dev_off=0, buf_sz=0;

	if( 0==(gCap.capabilities&V4L2_CAP_STREAMING) ) {
		printf("No memory mapping supported !\n");
		exit(1);
	}

	gMapSize=req_buffer(vbifd, &dev_off, &gBufferCount, &buf_sz);
	if( gMapSize ) {
		printf("Mapping from device %d\n", dev_off);
		if( buf_sz!=gBufferSize ) {
			printf("weird !!\n");
			gBufferSize = buf_sz;
		}

		gVAMap = (int)mmap(0, gMapSize, PROT_READ|PROT_WRITE,
				MAP_SHARED, vbifd, (off_t)dev_off);
		if( MAP_FAILED==(void *)gVAMap ) {
			perror("mmap");
			exit(1);
		}
	} 
	else 
		exit(1);

	return;
}

int main(int argc,char **argv)
{
	int vbifd;

	get_param(argc, argv);

	if ((vbifd = open(vbi_name, O_RDWR)) < 0) {	
		perror(vbi_name);
		exit(1);
	}
	
	if( !query_fmt(vbifd) ) return 1;
	init_win();
	buffer_alloc(vbifd);

	buffer_q(vbifd);
	dma_go(vbifd);

	close(vbifd);
	return 0;
}
