/*
 *  xtv.c
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  ========================================================================
 *  
 *  xtv.c is a video player tool to test V4L2 driver
 *
 *  Compile
 *	gcc -O2 -o xtv xtv.c /usr/X11R6/lib/libX11.so /usr/X11R6/lib/libXext.so
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <asm/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <unistd.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#include "videodev2.h"
#include "error.h"

/*
 * Change these parameters to change setting
 */
#define		BUF_FORMAT_TYPE		(V4L2_BUF_TYPE_VIDEO_CAPTURE)
#define		MAX_BUFFER	16

// to query X-Windows color depth and format
void output_X_query(void);
int output_X_error(Display * d, XErrorEvent * xeev);
int output_X_ioerror(Display *d);
int output_X_handle();
void output_X_start();
void output_X_stop(void);
int output_X_write(int fbnr, int picnr);
void output_X_abort(int x);
int file_desc;
struct v4l2_format fmt;

#define OUTPUT_X_EVENTMASK (KeyPressMask)

// to keep what format we want
static int output_pix_fmt;

static int output_width;
static int output_height;
static int output_bytes = 0; //w*h, a frame
char *output_ptrs[MAX_BUFFER];
static struct v4l2_fmtdesc gFmtDesc;
int q_flag = 0;
char error_progname[] = "xTV";
char error_text_buf[1024];
char *error_text = error_text_buf;
inline void error_abort(int c) { fprintf(stderr, "%s", error_text); exit(c); }

char *output_X_buffer = NULL;
XShmSegmentInfo output_X_shminfo;
Display *output_X_disp = NULL;
Window output_X_win = 0;
XImage *output_X_ximage = NULL;
XImage *output_X_logo = NULL;
GC output_X_gc = 0;
int output_X_errorcnt = 0;
int output_X_noerr = 0;
Colormap output_X_colormap = 0;
static Atom output_X_WM_PROTOCOLS = 0, output_X_WM_DELETE_WINDOW = 0;

static int screen_num;
static XFontStruct* font_info;

void select_X_font(Display *display, GC gc, char *font_name);
void draw_msg(char *s);
void process_code(KeySym keysym);

//static struct v4l2_capability gCap;
static char my_video_dev[256] = "/dev/video0";
static char *gVAMap;		//VA we map device
static int gMapSize;		//size we map device
static int chnls = 0;
static int ch_cur = 0;
static int need_pause = 0;
static int entry_process = 0;
static char entry_buff[256];
static int entry_pos = 0;
static int gD4_width = 0;	//yet set
static int gD4_height = 0;
struct cap_id_str {
	char name[32];
	unsigned int id;
};


/*
 * VIDIOC_ENUMINPUT
 * VIDIOC_ENUM_FMT
 * 1. Enumerate input and set desired input
 * NOTE: if input is tuner, set desired frequency as well
 * 2. Enumerate image format and remember desired format
 * 3. Set standard
 * Out	: gFmtDesc.index - desired image format
 * Ret	: 0 - fail
 *	  otherwise - success
 */
int show_info(int fd)
{
	int ret, idx;		//if none matches, set input 0
	struct v4l2_fmtdesc fmt;

	gFmtDesc.index = -1;

	/* Output Format */
        for( idx=0; 1; idx++ ) {
                fmt.index = idx;
                fmt.type = BUF_FORMAT_TYPE;
                if((ret = ioctl(fd, VIDIOC_ENUM_FMT, &fmt)))
                	break;
                if( output_pix_fmt == fmt.pixelformat ) {
                        gFmtDesc = fmt;
			break;
                } 
        }

	return 1;
}

/*
 * VIDIOC_S_FMT, VIDIOC_G_FMT
 * Set desired format, dimension (maximum) and interlaced frame
 * Out	: output_width - output frame width
 *	  output_height - output frame height
 *	  output_bytes - size of output frame
 * Ret	: 0 - fail
 *	  otherwise - success
 */
int set_format(int fd)
{
	v4l2_std_id std;
	int ret;

	// set dimension from std
	ret = ioctl(fd, VIDIOC_G_STD, &std);
	if(ret < 0) {
		perror("VIDIOC_G_STD");
		return 0;
	}

	gD4_width = 720;
	if( (std & V4L2_STD_NTSC_M) ||
	    (std & V4L2_STD_NTSC_M_JP) ||
	    (std & V4L2_STD_PAL_Nc) ) {
		gD4_height = 480;
	}
	else {
		gD4_height = 576;
	}

	if( -1==gFmtDesc.index ) {
		printf("This format is not supported\n");
		return 0;
	}

	fmt.type = BUF_FORMAT_TYPE;
	ret = ioctl(fd, VIDIOC_G_FMT, &fmt);
	fmt.type = BUF_FORMAT_TYPE;
	fmt.fmt.pix.pixelformat = gFmtDesc.pixelformat;
	fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;
	fmt.fmt.pix.bytesperline = 0;
	
	fmt.fmt.pix.width = gD4_width;
	fmt.fmt.pix.height = gD4_height;
	
	ret = ioctl(fd, VIDIOC_S_FMT, &fmt);
	if( ret ) {
		perror("VIDIOC_S_FMT");
		return 0;
	}

	output_width = fmt.fmt.pix.width;
	output_height = fmt.fmt.pix.height;
	output_bytes = output_height * fmt.fmt.pix.bytesperline;

	return 1;
}

/*
 * VIDIOC_REQBUFS
 * VIDIOC_QUERYBUF
 * Allocate buffer of device but return the length for
 * mapping at once, which must be continuous
 * In	: fd - dev file
 *	  count - number of buffer requested
 * Out	: count - number of buffer allocated
 *	  dev_off - from where (of dev) we start mapping
 * Ret	: 0 - fail
 *	  otherwise - length to map from offset (of device)
 */
int allocate_buffer(int fd, int *dev_off, int *count, int *buf_sz)
{
	struct v4l2_requestbuffers reqbuf;
	struct v4l2_buffer buf;
	int ret, i;
	int offset = -1, size;

	memset(&reqbuf, 0, sizeof(reqbuf));
	reqbuf.type = BUF_FORMAT_TYPE;
	reqbuf.memory = V4L2_MEMORY_MMAP;
	reqbuf.count = *count;

	ret = ioctl(fd, VIDIOC_REQBUFS, &reqbuf);
	if( ret ) {
		perror("VIDIOC_REQBUFS");
		return 0;
	} 

	for( i=0, size=0, *count=0; i<reqbuf.count; i++ ) {
		memset(&buf, 0, sizeof(buf));
		buf.type = reqbuf.type;
		buf.index = i;
		buf.memory = reqbuf.memory;
		ret = ioctl(fd, VIDIOC_QUERYBUF, &buf);
		if( ret ) {
			perror("VIDIOC_QUERYBUF");
			return 0;
		}
		*buf_sz = buf.length;

		if( 0==size ) offset = *dev_off = buf.m.offset;
		if( offset==buf.m.offset ) {
			offset += buf.length;
			size += buf.length;
			*count = *count + 1;
		} else {
			/* not continuous*/
			printf("Buffers not all continuous\n");
		}
	}

	return size;
}

/*
 * Negotiate with driver and set image size, format, video standard,
 * frequency ...
 * Ret	: 0 - fail
 *	  otherwise success
 */
int driver_nego(int *fd)
{
	fprintf(stderr, "open %s: ", my_video_dev);
	if( -1==(*fd=open(my_video_dev, O_RDWR)) ) {
		fprintf(stderr, "%s\n", strerror(errno));
		return 0;
	}
	else {
		fprintf(stderr, "Success\n");
	}

	if( show_info(*fd) && set_format(*fd) ) {
		fcntl(*fd, F_SETFD, FD_CLOEXEC);
		return 1;
	}

	return 0;
}

/*
 * Map device to gVAMap and queue all the buffers
 * Ret	: 0 - fail
 *	  otherwise - number of buffer queued(mapped)
 */
int do_map_n_q(int fd)
{
	char *map_start;
	int dev_start, buf_size, buf_count=MAX_BUFFER;
	int i;
	struct v4l2_buffer buf;

	gMapSize = allocate_buffer(fd, &dev_start, &buf_count, &buf_size);
	if( 0==gMapSize ) {
		printf("zero size to map\n");
		return 0;
	}

	gVAMap = mmap(0, gMapSize, PROT_READ|PROT_WRITE, MAP_SHARED,
							fd, dev_start);
	if( MAP_FAILED==gVAMap ) {
		perror("mmap");
		return 0;
	}

	for( i=0, map_start=gVAMap; i<buf_count; i++, map_start+=buf_size ) {
		memset(&buf, 0, sizeof(buf));
		buf.index = i;
		buf.type = BUF_FORMAT_TYPE;
		buf.memory = V4L2_MEMORY_MMAP;
		if( ioctl(fd, VIDIOC_QBUF, &buf) ) {
			perror("VIDIOC_QBUF");
			munmap(gVAMap, gMapSize);
			return 0;
		}
		output_ptrs[i] = map_start;
	}

	return buf_count;
}

void usage()
{
	printf("Usage: xtv \n\n");
	printf("\t   video-dev : Default to \"/dev/video0\"\n");
	printf("\n");
	exit(1);
}

static void get_param(int argc, char *argv[])
{
	while(1) {
		int arg;

		if( -1 == (arg = getopt(argc, argv, "?h")) )
			break;

		switch(arg) {
		case '?':
		case 'h':
			usage();
			exit(0);
		}
	}

	// non-options are treated as device name
	if(optind < argc)
		strncpy(my_video_dev, argv[optind], sizeof(my_video_dev));
}

int main(int argc, char *argv[])
{
	int i, j;
	int buf_count;
	struct v4l2_buffer buf;

	entry_buff[0] = '\0';

	/*
	 * to query current X-Windows color depth and format first,
	 * then to set the device output format
	 */
        output_X_query();

	get_param(argc, argv);
	if( !driver_nego(&file_desc) ) return 0; // Set default video input, standard and channel
	buf_count = do_map_n_q(file_desc);
	if( 0==buf_count ) return 0;

	if( ioctl(file_desc, VIDIOC_STREAMON, &i) ) goto err;
	output_X_start();

	i = 1;
	while( !q_flag ) {
		for( j=0; j<buf_count; j++ ) {
			memset(&buf, 0, sizeof(buf));
			buf.type = BUF_FORMAT_TYPE;
			buf.memory = V4L2_MEMORY_MMAP;
			if( ioctl(file_desc, VIDIOC_DQBUF, &buf) ) {
				perror("VIDIOC_DQBUF");
				goto err1;
			}
			if( 1==output_X_write(j, i++) ) break;	//because restreaming

			if( ioctl(file_desc, VIDIOC_QBUF, &buf) ) {
				perror("VIDIOC_QBUF");
				goto err1;
			}
		}
		
		if( need_pause ) {
			printf("need_pause\n");
			while( need_pause ) output_X_handle();
		}
	}
err1:
	printf("Stop\n");
	if( ioctl(file_desc, VIDIOC_STREAMOFF, &i) )
		perror("VIDIOC_STREAMOFF");
err:
	output_X_stop();
	munmap(gVAMap, gMapSize);
	close(file_desc);
	return 0;
}

void draw_msg(char *s)
{
	XSetForeground(output_X_disp, output_X_gc, XBlackPixel(output_X_disp, screen_num));
	XFillRectangle(output_X_disp, output_X_win, output_X_gc, 0, 480, 400, 496);
	XSetForeground(output_X_disp, output_X_gc, XWhitePixel(output_X_disp, screen_num));
	XDrawString(output_X_disp, output_X_win, output_X_gc, 10, 492, s, strlen(s));
}

void process_code(KeySym keysym)
{

	switch (keysym)
	{
		case XK_Return:
			entry_process = 0;
			break;
	
		case XK_BackSpace:
			entry_pos--;
			if(entry_pos < 0)
				entry_pos = 0;
			
			entry_buff[entry_pos] = '\0';
			draw_msg(entry_buff);
			break;
	
		case XK_Delete:
			break;
	
		default:
			printf("%c pressed\n", (int)keysym);
			entry_buff[entry_pos++] = keysym;
			entry_buff[entry_pos] = '\0';
			printf("entry: %s\n", entry_buff);
			draw_msg(entry_buff);
			break;
	}

}

////////////////////////////////////////////////////////////////////

int output_X_error(Display * d, XErrorEvent * xeev)
{
	output_X_errorcnt++;
	return 0;
}

int output_X_ioerror(Display *d)
{
	if (!output_X_noerr) {
		ERROR(MSG("fatal X Windows error"));
		END(1);
	}
	return 0;
}

/*
 * Press 'Q' or 'q' to stop
 *       'P' or 'p' to pause display
 * Ret	: 0 - ok
 *	  1 - toggle standard and re-streamon
 */
int output_X_handle()
{
	XEvent xev;
	KeySym keysym;

	while (XCheckWindowEvent(output_X_disp, output_X_win, OUTPUT_X_EVENTMASK, &xev)) {
		switch (xev.type)
		{
		case KeyPress:
			keysym = XKeycodeToKeysym(output_X_disp, xev.xkey.keycode, 0);
			
			if(entry_process)
			{
				process_code(keysym);
			}
			else if (keysym == XK_Q || keysym == XK_q)
			{
				printf("q_flag on\n");
				q_flag = 1;
			}
			else
			{
				if((keysym == XK_comma) || (keysym == XK_period))
				{
				int ch =  ch_cur;
				
					if(keysym == XK_comma)
					{
						ch--;
						if(ch < 0)
							ch = chnls - 1;
					}
					else if(keysym== XK_period)
					{
						ch++;
						if(ch > (chnls - 1))
							ch = 0;
					}
					
				}
				if (keysym == XK_P || keysym == XK_p)
				{
					if(need_pause)
						need_pause = 0;
					else
						need_pause = 1;
				}
				if( keysym == XK_T || keysym == XK_t ) {
		v4l2_std_id std;
		int ret;

		/*
		 * toggle between NTSC_M and PAL_B
		 * stop streaming
		 */
		if( ioctl(file_desc, VIDIOC_STREAMOFF, &std) ) {
			perror("VIDIOC_STREAMOFF");
			exit(1);
		}
		munmap(gVAMap, gMapSize);

		if( 480==output_height ) {
			fmt.fmt.pix.height = 
			output_height = 576;
			std = V4L2_STD_PAL_B;
		} else {
			fmt.fmt.pix.height = 
			output_height = 480;
			std = V4L2_STD_NTSC_M;
		}

		ret = ioctl(file_desc, VIDIOC_G_STD, &std);
		if(ret < 0) {
			perror("VIDIOC_G_STD");
			exit(1);
		}

		ret = ioctl(file_desc, VIDIOC_S_FMT, &fmt);	//not necessary
		if( ret ) {
			perror("VIDIOC_S_FMT");
			exit(1);
		}

		if( 0==do_map_n_q(file_desc) ) exit(1);

		if( ioctl(file_desc, VIDIOC_STREAMON, &std) ) {
			perror("VIDIOC_STREAMON");
			exit(1);
		}
		printf("Restart stream\n");
		return 1;
				} //if( keysym == XK_T || keysym == XK_t )
			}
			break;

		case ClientMessage:
			if (xev.xclient.message_type == output_X_WM_PROTOCOLS && xev.xclient.data.l[0] == output_X_WM_DELETE_WINDOW) {
				ERROR(NOERR("user abort"));
				END(0);
			}
			break;

		default:
			break;

		}
	} 
	return 0;
}


/////////////////////////////////////////////////////////////
// output_X_query()
// 1. open X windows display
// 2. query the depth of screen and red_mask information,
//    to know exactly color format
// 3. then parsing these information, to determine what
//    format should V4L2 video capture device should
//    output
/////////////////////////////////////////////////////////////
void output_X_query()
{
	int depth;
	unsigned long red_mask;
	
	ASSERT((output_X_disp = XOpenDisplay(NULL)),(MSG("open display failed\n")));
        
        depth = DefaultDepthOfScreen(DefaultScreenOfDisplay(output_X_disp));
 	red_mask = 
 	    DefaultVisualOfScreen(DefaultScreenOfDisplay(output_X_disp))->red_mask;
  
 	switch(depth)
 	{
 	    case 16:
 	        switch(red_mask)
 	        {
 	            case 0x0000f800:
 	            	output_pix_fmt = V4L2_PIX_FMT_RGB565;
 	            	break;
 	            case 0x0000001f:
 	            	output_pix_fmt = V4L2_PIX_FMT_RGB565X;
 	            	break;
 	            default:
 	                    ERROR(MSG("open display failed: color format not support\n"));
		            output_X_stop();
		            END(-1);
 	            	break;
 	        }
 	    	break;
 	    case 24:
 	        switch(red_mask)
 	        {
 	            case 0x00FF0000:
 	                output_pix_fmt = V4L2_PIX_FMT_BGR32;
 	            	break;
 	            case 0x000000FF:
 	                output_pix_fmt = V4L2_PIX_FMT_RGB32;
 	            	break;
 	            default:
                            ERROR(MSG("open display failed: color format not support\n"));
		            output_X_stop();
		            END(-1); 	            
 	            	break;
 	        }
 	    	break;
 	    default:
                    ERROR(MSG("open display failed: color format not support\n"));
                    output_X_stop();
                    END(-1); 	    
 	    	break;
 	}
}

/*
 * In	: output_width, output_height
 */
void output_X_start()
{
	static XGCValues values;
	XSizeHints hints;
	XWMHints wmhints;
	XTextProperty windowName, iconName;
	Window root;
	int depth;
	char *name = "xTV";

	output_X_shminfo.shmaddr = 0;
	output_X_shminfo.shmid = -1;

	ASSERT(signal(SIGPIPE,output_X_abort) != SIG_ERR,(ERRMSG("signal")));

	output_X_noerr = 0;
	output_X_errorcnt = 0;
	XSetErrorHandler(output_X_error);
	XSetIOErrorHandler(output_X_ioerror);
	
	screen_num = DefaultScreen(output_X_disp);
	
	root = DefaultRootWindow(output_X_disp);
	output_X_win = XCreateSimpleWindow(output_X_disp, root, 0, 0,
				output_width, output_height, 0,
				XWhitePixel(output_X_disp,screen_num),
				XBlackPixel(output_X_disp,screen_num));

	ASSERT(output_X_win,(MSG("create window failed\n")));
	
	hints.flags = PSize|PMaxSize|PMinSize;
	hints.min_width = hints.max_width = hints.width = output_width;
	hints.min_height = hints.max_height = hints.height = output_height;

	wmhints.input = True;
	wmhints.flags = InputHint;
	
	XStringListToTextProperty(&name, 1 ,&windowName);
	XStringListToTextProperty(&name, 1 ,&iconName);

	XSetWMProperties(output_X_disp, output_X_win, &windowName, &iconName, NULL, 0, &hints, &wmhints, NULL);
	
	output_X_WM_DELETE_WINDOW = XInternAtom(output_X_disp, "WM_DELETE_WINDOW", False);
	output_X_WM_PROTOCOLS = XInternAtom(output_X_disp, "WM_PROTOCOLS", False);
	XSetWMProtocols(output_X_disp, output_X_win, &output_X_WM_DELETE_WINDOW, 1);

	XSelectInput(output_X_disp, output_X_win, OUTPUT_X_EVENTMASK);

	XMapRaised(output_X_disp, output_X_win);
	
	output_X_gc = XCreateGC(output_X_disp, output_X_win, 0, &values);
	
	select_X_font(output_X_disp, output_X_gc, "-misc-fixed-medium-r-normal-*-*-140-*-*-c-*-*");

	XSetForeground(output_X_disp, output_X_gc, WhitePixel(output_X_disp,screen_num));
	
	XSync(output_X_disp, 0);
	
	depth = DefaultDepthOfScreen(DefaultScreenOfDisplay(output_X_disp));

	output_X_ximage = XShmCreateImage(output_X_disp, DefaultVisual(output_X_disp, screen_num), 
					  depth, ZPixmap, NULL, &output_X_shminfo, output_width, output_height);
	ASSERT(output_X_ximage,(MSG("CreateImage Failed\n")));

	output_X_shminfo.shmid = shmget(IPC_PRIVATE, output_X_ximage->bytes_per_line*output_X_ximage->height, IPC_CREAT|0777);
	ASSERT(output_X_shminfo.shmid >= 0,(ERRMSG("shmget failed")));
	output_X_buffer = output_X_ximage->data = output_X_shminfo.shmaddr = shmat(output_X_shminfo.shmid, 0, 0);
	
	XShmAttach(output_X_disp, &output_X_shminfo);
	
	XSync(output_X_disp, 0);
	
	if (output_X_errorcnt) {
		fprintf(stderr, "Shared memory unavailable, using regular images\n");
		shmdt(output_X_shminfo.shmaddr);
		output_X_shminfo.shmaddr = 0;
		
		SAFE_MALLOC(output_X_buffer,((depth + 7) / 8) * output_width * output_height);
		output_X_ximage = XCreateImage(output_X_disp, DefaultVisual(output_X_disp, screen_num), 
					       depth, ZPixmap, 0, output_X_buffer, output_width, output_height, depth,
					       output_width * ((depth + 7) / 8));
		ASSERT(output_X_ximage,(MSG("CreateImage Failed\n")));
	}
	XFlush(output_X_disp);
}

void output_X_stop(void)
{
	output_X_noerr = 1;
	if (output_X_colormap) {
		XFreeColormap(output_X_disp,output_X_colormap);
		output_X_colormap = 0;
	}
	if (output_X_ximage) XDestroyImage(output_X_ximage);
	
	if (output_X_shminfo.shmid >= 0) {
		XShmDetach(output_X_disp, &output_X_shminfo);
		if (output_X_shminfo.shmaddr) {
			shmdt(output_X_shminfo.shmaddr);
			output_X_buffer = NULL;
		}
		shmctl(output_X_shminfo.shmid, IPC_RMID, 0);
		output_X_shminfo.shmid = -1;
	}
	if (output_X_buffer) {
		SAFE_FREE(output_X_buffer);
		output_X_buffer = NULL;
	}

	if (output_X_win) {
		XEvent xev;
		if (output_X_gc) {
			XFreeGC(output_X_disp, output_X_gc);
			output_X_gc = 0;	
		}
		XUnmapWindow(output_X_disp,output_X_win);
		XFlush(output_X_disp);
		while (XCheckWindowEvent(output_X_disp, output_X_win, OUTPUT_X_EVENTMASK, &xev));
		XDestroyWindow(output_X_disp,output_X_win);
		output_X_win = 0;
	}
	if (output_X_colormap) {
		XFreeColormap(output_X_disp,output_X_colormap);
		output_X_colormap = 0;
	}
	if (output_X_disp) {
		XCloseDisplay(output_X_disp);
		output_X_disp = NULL;
	}
}


/*
 * Ret	: 0 - ok
 *	  1 - toggle standard and re-streamon
 */
int output_X_write(int fbnr, int picnr)
{

	memcpy(output_X_buffer, output_ptrs[fbnr], output_bytes);
	
	if (output_X_shminfo.shmaddr)
		XShmPutImage(output_X_disp, output_X_win, output_X_gc, output_X_ximage, 0, 0, 0, 0, output_width, output_height, False);
	else
		XPutImage(output_X_disp, output_X_win, output_X_gc, output_X_ximage,  0, 0, 0, 0, output_width, output_height);
	
	XFlush(output_X_disp);
	return output_X_handle();
}

void output_X_abort(int x)
{
}


void select_X_font(Display *display, GC gc, char *font_name)
{

	font_info = XLoadQueryFont(display, font_name);
	if (!font_info)
	{
		fprintf(stderr, "XLoadQueryFont: failed loading font '%s'\n", font_name);
	}

	/* assign the given font to our GC. */
	XSetFont(display, gc, font_info->fid);
}

